import argparse
import os
from functools import *
from glob import glob

import numpy as np
import SimpleITK as sitk
from rich import print
from rich.progress import track
from scipy.linalg import hadamard

from asltk.utils import load_image, save_image

parser = argparse.ArgumentParser(
    prog='Generate Subtracted ASL Image',
    description='Python script to assist in reconstructing the ASL image already subtract from control and tagged volumes. This script assumes that the ASL raw data was acquired using a MRI imaging protocols based on Hadamard matrix acquisition. There are some default values for the PLD and LD, but the user can inform the values used in the MRI protocol. Please, be aware about the default values and inform the correct values used in the MRI protocol.',
)
parser._action_groups.pop()
required = parser.add_argument_group(title='Required parameters')
optional = parser.add_argument_group(title='Optional parameters')


required.add_argument(
    'datafolder',
    type=str,
    help='Folder containing the ASL raw data obtained from the MRI scanner. This folder must have the Nifti files converted from the DICOM files.  By default the output file name adopted is pcasl.(file_fmt), where file_fmt is the file format informed in the parameter --file_fmt. TIP: One can use other tools such dcm2nii to convert DICOM data to Nifti.',
)
required.add_argument(
    '--matrix_order',
    type=int,
    required=False,
    default=8,
    help='Informs the Hadamar matrix size used in the MRI imaging protocol. This must be a positive power-of-two integer (n^2).',
)
required.add_argument(
    '--dynamic_vols',
    type=int,
    required=False,
    default=2,
    help='Informs the number of dynamic volumes used in the MRI acquisition.',
)
required.add_argument(
    '--pld',
    type=str,
    nargs='+',
    required=False,
    default=[170.0, 270.0, 370.0, 520.0, 670.0, 1070.0, 1870.0],
    help='Posts Labeling Delay (PLD) trend, arranged in a sequence of float numbers. If not passed, the default values will be used.',
)
required.add_argument(
    '--ld',
    type=str,
    nargs='+',
    required=False,
    default=[100.0, 100.0, 150.0, 150.0, 400.0, 800.0, 1800.0],
    help='Labeling Duration trend (LD), arranged in a sequence of float numbers. If not passed, the default values will be used.',
)
optional.add_argument(
    '--output_folder',
    type=str,
    nargs='?',
    default='',
    help='The output folder that is the reference to save the output image. By default, the output image will be saved in the same folder as the input data. If informed, the output image will be saved in the folder informed.',
)
optional.add_argument(
    '--mask',
    type=str,
    nargs='?',
    default='',
    help='Image mask defining the ROI where the calculations must be done. Any pixel value different from zero will be assumed as the ROI area. Outside the mask (value=0) will be ignored. If not provided, the entire image space will be calculated.',
)
optional.add_argument(
    '--te',
    type=float,
    nargs='+',
    default=[13.56, 67.82, 122.08, 176.33, 230.59, 284.84, 339.1, 393.36],
    help='Time of Echos (TE), arranged in a sequence of float numbers. This is only required for multi-TE ASL data. This sequence of values must be in accordance with the number of volumes acquired in the MRI protocol.',
)
optional.add_argument(
    '--dw',
    type=float,
    nargs='+',
    default=[0, 50, 100, 250],
    help='Diffusion weights (DW), arranged in a sequence of float numbers. This is only required for multi-DW ASL data. This sequence of values must be in accordance with the number of volumes acquired in the MRI protocol.',
)
optional.add_argument(
    '--file_fmt',
    type=str,
    required=False,
    default='nii',
    help='Define the file format to load the ASL data in the datafolder parameter and also be used for saving the output image. The default is Nifti format (nii). File formats allowed: nii, mha, nrrd. TIP: This file format depends on the output of the DICOM converter tool used, then it is important to check the output format of the tool used to convert the DICOM files to Nifti files.',
)
optional.add_argument(
    '--verbose',
    action='store_true',
    help='Show more details thoughout the processing.',
)

args = parser.parse_args()

# Check input parameters
def checkUpParameters():
    is_ok = True

    # Check if the data folder exists
    if not os.path.exists(args.datafolder):
        print(
            f"[red]Error: The data folder '{args.datafolder}' does not exist.[/red]"
        )
        is_ok = False

    # Check if the matrix order is a power-of-two integer
    if not (args.matrix_order & (args.matrix_order - 1) == 0):
        print(
            f'[red]Error: The matrix order must be a power-of-two integer.[/red]'
        )
        is_ok = False

    # Check if the dynamic volumes is a positive integer
    if args.dynamic_vols <= 0:
        print(
            f'[red]Error: The dynamic volumes must be a positive integer.[/red]'
        )
        is_ok = False

    # Check if file format is allowed
    if args.file_fmt not in ['nii', 'mha', 'nrrd']:
        print(
            f"[red]Error: The file format '{args.file_fmt}' is not allowed.[/red]"
        )
        is_ok = False

    # Check if mask file exists, if informed
    if args.mask and not os.path.exists(args.mask):
        print(f"[red]Error: The mask file '{args.mask}' does not exist.[/red]")
        is_ok = False

    # If passed, check if the output folder exists
    if args.output_folder and not os.path.exists(args.output_folder):
        print(
            f"[red]Error: The output folder '{args.output_folder}' does not exist.[/red]"
        )
        is_ok = False

    return is_ok


def load_asl_data(datafolder, file_fmt):
    asl_files = [
        os.path.join(datafolder, f)
        for f in os.listdir(datafolder)
        if f.endswith(f'.{file_fmt}')
    ]
    asl_images = [load_image(f) for f in asl_files]
    return asl_images


# Start the script execution
if not checkUpParameters():
    raise RuntimeError(
        'One or more arguments are not well defined. Please, revise the script call.'
    )

if args.verbose:
    print(' --- Script Input Data ---')
    print('Required parameters:')
    print('Input data folder path: ' + args.datafolder)
    print('Matrix order: ' + str(args.matrix_order))
    print('Dynamic volumes: ' + str(args.dynamic_vols))
    print('PLD: ' + str(args.pld))
    print('LD: ' + str(args.ld))
    print('----------------')
    print('Optional parameters:')
    if args.output_folder:
        print('Output folder path: ' + str(args.output_folder))
    else:
        print('Output folder path: assuming the input folder path')
    if args.mask:
        print('Mask image file path: ' + str(args.mask))
    else:
        print('Mask image file path: not applying brain mask')
    print('TE values: ' + str(args.te))
    print('DW values: ' + str(args.dw))
    print('Input/Output file format: ' + args.file_fmt)


# Load all the Nifti files in the folder
asl_images = load_asl_data(args.datafolder, args.file_fmt)
# If mask is passed, check if shape is the same as the ASL images
if args.mask:
    mask_image = load_image(args.mask)
    if mask_image.shape != asl_images[0].shape[1:]:
        raise ValueError(
            'The mask image shape does not match the ASL images shape.'
        )
    asl_images = [asl_image * (mask_image > 0) for asl_image in asl_images]

# Load the input ASL data (PLD, LD, and TE, DWI if available)
try:
    pld = [float(s) for s in args.pld]
    ld = [float(s) for s in args.ld]
    te = [float(s) for s in args.te]
    dw = [float(s) for s in args.dw]
except:
    pld = [float(s) for s in str(args.pld[0]).split()]
    ld = [float(s) for s in str(args.ld[0]).split()]
    te = [float(s) for s in str(args.te[0]).split()]
    dw = [float(s) for s in str(args.dw[0]).split()]

# Process the data using Hadamar's product
boli_cnt = args.matrix_order - 1  # number of sub-bolis
nTe = len(te)  # number of echo times
# TODO Adapt the script for DW images as well

h_mtrx = hadamard(args.matrix_order)
subtr_mtrx = -h_mtrx[:, 1 : args.matrix_order]
fn = [
    f
    for f in glob(os.path.join(args.datafolder, f'*.{args.file_fmt}'))
    if 'pcasl' not in os.path.basename(f)
]
head = sitk.ReadImage(fn[0])

for cnt in track(range(nTe), description='Processing volumes...'):
    dat = sitk.ReadImage(fn[cnt])
    dat_array = sitk.GetArrayFromImage(dat)
    dims = dat_array.shape

    if cnt == 0:
        dat_final = np.zeros(
            (nTe, boli_cnt, dims[1], dims[2], dims[3]), dtype=np.float64
        )

    dat_aux = np.zeros(
        (
            args.dynamic_vols,
            dims[0] // args.dynamic_vols,
            dims[1],
            dims[2],
            dims[3],
        ),
        dtype=np.float64,
    )

    dat_aux[0, ...] = dat_array[::2, ...]
    dat_aux[1, ...] = dat_array[1::2, ...]

    unsubtr_data_mean = np.mean(dat_aux, axis=0)

    Subtr_phases = np.zeros(unsubtr_data_mean[1:, ...].shape, dtype=np.float64)

    for bolus in range(boli_cnt):
        vector = subtr_mtrx[:, bolus]
        for line in range(args.matrix_order):
            Subtr_phases[boli_cnt - bolus - 1, ...] += (
                vector[line] * unsubtr_data_mean[line, ...]
            )

    dat_final[cnt, ...] = Subtr_phases * float(
        dat.GetMetaData('scl_slope')
    ) + float(dat.GetMetaData('scl_inter'))

# Create the final NIfTI image
dat_final_sitk = sitk.GetImageFromArray(dat_final)
# TODO Set the image information from the input ASL header
# dat_final_sitk.SetSpacing(head.GetSpacing())
# dat_final_sitk.SetDirection(head.GetDirection())
# dat_final_sitk.SetOrigin(head.GetOrigin())
# sitk.WriteImage(dat_final_sitk, os.path.join(args.datafolder, "pcasl.nii.gz"))

# Save the output images in the output folder (or other path passed by the user)
output_path = os.path.join(
    args.output_folder if args.output_folder else args.datafolder,
    f'pcasl.{args.file_fmt}',
)
if args.verbose:
    print('Saving file at: ' + output_path)
save_image(dat_final, output_path)

if args.verbose:
    print('Execution: ' + parser.prog + ' finished successfully!')



from multiprocessing import Array, Pool, cpu_count

import numpy as np
from rich import print
from rich.progress import Progress
from scipy.optimize import curve_fit

from asltk.asldata import ASLData
from asltk.aux_methods import _apply_smoothing_to_maps, _check_mask_values
from asltk.logging_config import get_logger, log_processing_step
from asltk.mri_parameters import MRIParameters

# Global variables for multiprocessing
t2_map_shared = None
brain_mask = None
data = None
TEs = None


class T2Scalar_ASLMapping(MRIParameters):
    """
    Class for voxel-wise T2 mapping from multi-echo ASL data.

    This class provides methods to calculate T2 relaxation maps from multi-echo ASL MRI data.
    It supports brain masking, multiprocessing for fast computation, and optional smoothing.

    Main methods:
        - set_brain_mask: Set a binary mask to restrict T2 fitting to brain voxels.
        - create_map: Compute T2 maps using multiprocessing (output shape: (N_PLDS, Z, Y, X)).
        - get_t2_maps: Retrieve the computed T2 maps.
        - get_mean_t2s: Retrieve mean T2 values per PLD.
    """

    def __init__(self, asl_data: ASLData) -> None:
        super().__init__()
        self._asl_data = asl_data
        self._te_values = self._asl_data.get_te()
        self._pld_values = self._asl_data.get_pld()

        # Check if the ASLData has TE and PLD values
        if self._te_values is None or not self._pld_values:
            raise ValueError('ASLData must provide TE and PLD values.')

        # Check if the ASLData has DW values (not allowed for T2 mapping)
        if self._asl_data.get_dw() is not None:
            raise ValueError('ASLData must not include DW values.')

        self._brain_mask = np.ones(self._asl_data('m0').shape)
        self._t2_maps = None  # Will be 4D: (N_PLDS, Z, Y, X)
        self._mean_t2s = None

    def set_brain_mask(self, brain_mask: np.ndarray, label: int = 1):
        """
        Set a brain mask to restrict T2 fitting to specific voxels.

        Args:
            brain_mask (np.ndarray): Binary or integer mask with the same shape as the M0 image. Nonzero values indicate voxels to include.
            label (int, optional): The label value to use as foreground (default: 1).

        The mask should be a 3D numpy array matching the spatial dimensions of the ASL data.
        """
        _check_mask_values(brain_mask, label, self._asl_data('m0').shape)
        binary_mask = (brain_mask == label).astype(np.uint8) * label
        self._brain_mask = binary_mask

    def get_t2_maps(self):
        """Get the T2 maps storaged at the T2Scalar_ASLMapping object

        Returns:
            (np.ndarray): The T2 maps that is storaged in the
            T2Scalar_ASLMapping object
        """
        return self._t2_maps

    def get_mean_t2s(self):
        """Get the mean T2 values calculated from the T2 maps

        Returns:
            (list): The mean T2 values for each PLD
        """
        return self._mean_t2s

    def create_map(
        self, cores=cpu_count(), smoothing=None, smoothing_params=None
    ):
        """
        Compute T2 maps using multi-echo ASL data and a brain mask, with multiprocessing.

        This method uses multiprocessing to accelerate voxel-wise T2 fitting. The output is a 4D array with shape (N_PLDS, Z, Y, X).

        Warning:
            For large datasets, memory usage can be significant due to parallel processing and storage of intermediate arrays.

        Args:
            cores (int, optional): Number of CPU cores for processing. Defaults to all available.
            smoothing (str, optional): Smoothing type ('gaussian', 'median', or None).
            smoothing_params (dict, optional): Smoothing parameters.

        Returns:
            dict: Dictionary with T2 maps ('t2', shape (N_PLDS, Z, Y, X)) and mean T2 values ('mean_t2').
        """
        logger = get_logger('t2_mapping')
        logger.info('Starting T2 map creation')

        data = self._asl_data('pcasl')
        mask = self._brain_mask
        TEs = np.array(self._te_values)
        PLDs = np.array(self._pld_values)
        n_tes, n_plds, z_axis, y_axis, x_axis = data.shape

        t2_maps_all = []
        mean_t2s = []

        for pld_idx in range(n_plds):
            logger.info(f'Processing PLD index {pld_idx} ({PLDs[pld_idx]} ms)')
            t2_map_shared = Array('d', z_axis * y_axis * x_axis, lock=False)
            log_processing_step(
                'Running voxel-wise T2 fitting',
                'this may take several minutes',
            )
            with Pool(
                processes=cores,
                initializer=_t2_init_globals,
                initargs=(t2_map_shared, mask, data, TEs),
            ) as pool:
                with Progress() as progress:
                    task = progress.add_task(
                        f'T2 fitting (PLD {PLDs[pld_idx]} ms)...', total=x_axis
                    )
                    results = [
                        pool.apply_async(
                            _t2_process_slice,
                            args=(i, x_axis, y_axis, z_axis, pld_idx),
                            callback=lambda _: progress.update(
                                task, advance=1
                            ),
                        )
                        for i in range(x_axis)
                    ]
                    for result in results:
                        result.wait()

            t2_map = np.frombuffer(t2_map_shared).reshape(
                z_axis, y_axis, x_axis
            )
            t2_maps_all.append(t2_map)
            mean_t2s.append(np.nanmean(t2_map))

        t2_maps_stacked = np.stack(
            t2_maps_all, axis=0
        )  # shape: (N_PLDS, Z, Y, X)
        self._t2_maps = t2_maps_stacked
        self._mean_t2s = mean_t2s

        logger.info('T2 mapping completed successfully')
        logger.info(
            f'T2 statistics - Mean: {np.mean(self._t2_maps):.4f}, Std: {np.std(self._t2_maps):.4f}'
        )

        output_maps = {
            't2': self._t2_maps,
            'mean_t2': self._mean_t2s,
        }

        return _apply_smoothing_to_maps(
            output_maps, smoothing, smoothing_params
        )


def _fit_voxel(signal, TEs):  # pragma: no cover
    """
    Fits a monoexponential decay model to the signal across TEs to estimate T2.

    Args:
        signal (np.ndarray): Signal intensities for different TEs.
        TEs (np.ndarray): Echo times (ms).

    Returns:
        float: Estimated T2 value (ms), or 0 if fitting fails.
    """

    def monoexp(te, S0, T2):
        return S0 * np.exp(-te / T2)

    # Check for valid signal
    if np.any(np.isnan(signal)) or np.max(signal) < 1:
        return 0

    try:
        popt, _ = curve_fit(
            monoexp, TEs, signal, p0=(np.max(signal), 80), bounds=(0, np.inf)
        )
        T2 = popt[1]
        if T2 <= 0 or np.isnan(T2):
            return 0
        return T2
    except Exception:
        return 0


def _t2_init_globals(t2_map_, brain_mask_, data_, TEs_):   # pragma: no cover
    global t2_map_shared, brain_mask, data, TEs
    t2_map_shared = t2_map_
    brain_mask = brain_mask_
    data = data_
    TEs = TEs_


def _t2_process_slice(i, x_axis, y_axis, z_axis, pld_idx):   # pragma: no cover
    for j in range(y_axis):
        for k in range(z_axis):
            if brain_mask[k, j, i]:
                signal = data[:, pld_idx, k, j, i]
                t2_value = _fit_voxel(signal, TEs)
                index = k * (y_axis * x_axis) + j * x_axis + i
                t2_map_shared[index] = t2_value
            else:
                index = k * (y_axis * x_axis) + j * x_axis + i
                t2_map_shared[index] = 0



import warnings
from typing import Any, Dict, Optional

import numpy as np

from asltk.smooth import isotropic_gaussian, isotropic_median


def _check_mask_values(mask, label, ref_shape):
    # Check wheter mask input is an numpy array
    if not isinstance(mask, np.ndarray):
        raise TypeError(f'mask is not an numpy array. Type {type(mask)}')

    # Check whether the mask provided is a binary image
    unique_values = np.unique(mask)
    if unique_values.size > 2:
        warnings.warn(
            'Mask image is not a binary image. Any value > 0 will be assumed as brain label.',
            UserWarning,
        )

    # Check whether the label value is found in the mask image
    label_ok = False
    for value in unique_values:
        if label == value:
            label_ok = True
            break
    if not label_ok:
        raise ValueError('Label value is not found in the mask provided.')

    # Check whether the dimensions between mask and input volume matches
    mask_shape = mask.shape
    if mask_shape != ref_shape:
        raise TypeError(
            f'Image mask dimension does not match with input 3D volume. Mask shape {mask_shape} not equal to {ref_shape}'
        )


def _apply_smoothing_to_maps(
    maps: Dict[str, np.ndarray],
    smoothing: Optional[str] = None,
    smoothing_params: Optional[Dict[str, Any]] = None,
) -> Dict[str, np.ndarray]:
    """Apply smoothing filter to all maps in the dictionary.

    This function applies the specified smoothing filter to all map arrays
    in the input dictionary. It preserves the original structure and only
    modifies the numpy arrays.

    Parameters
    ----------
    maps : dict
        Dictionary containing map arrays (e.g., {'cbf': array, 'att': array}).
    smoothing : str, optional
        Type of smoothing filter to apply. Options:
        - None: No smoothing (default)
        - 'gaussian': Gaussian smoothing using isotropic_gaussian
        - 'median': Median filtering using isotropic_median
    smoothing_params : dict, optional
        Parameters for the smoothing filter. Defaults depend on filter type:
        - For 'gaussian': {'sigma': 1.0}
        - For 'median': {'size': 3}

    Returns
    -------
    dict
        Dictionary with the same keys but smoothed arrays.

    Raises
    ------
    ValueError
        If smoothing type is not supported.
    """
    # Check it the smoothing_params is ok
    if smoothing_params is not None and not isinstance(smoothing_params, dict):
        raise TypeError(
            f'smoothing_params must be a dictionary. Type {type(smoothing_params)}'
        )
    if isinstance(smoothing_params, dict):
        if smoothing_params.get('size') or smoothing_params.get('sigma'):
            if smoothing_params.get('size') and not isinstance(
                smoothing_params['size'], int
            ):
                raise TypeError(
                    'Invalid smoothing parameter type. Size/Sigma must be an integer.'
                )
            if smoothing_params.get('sigma') and not isinstance(
                smoothing_params['sigma'], float
            ):
                raise TypeError(
                    'Invalid smoothing parameter type. Size/Sigma must be a float.'
                )

    if smoothing is None:
        return maps

    # Set default parameters
    if smoothing_params is None:
        if smoothing == 'gaussian':
            smoothing_params = {'sigma': 1.0}
        elif smoothing == 'median':
            smoothing_params = {'size': 3}
        else:
            smoothing_params = {}

    # Select smoothing function
    if smoothing == 'gaussian':
        smooth_func = isotropic_gaussian
    elif smoothing == 'median':
        smooth_func = isotropic_median
    else:
        raise ValueError(
            f'Unsupported smoothing type: {smoothing}. '
            "Supported types are: None, 'gaussian', 'median'"
        )

    # Apply smoothing to all maps
    smoothed_maps = {}
    for key, map_array in maps.items():
        if isinstance(map_array, np.ndarray):
            try:
                smoothed_maps[key] = smooth_func(map_array, **smoothing_params)
            except Exception as e:
                warnings.warn(
                    f'Failed to apply {smoothing} smoothing to {key} map: {e}. '
                    f'Using original map.',
                    UserWarning,
                )
                smoothed_maps[key] = map_array
        else:
            # Non-array values are passed through unchanged
            smoothed_maps[key] = map_array

    return smoothed_maps



import numpy as np
import SimpleITK as sitk


def rigid_body_registration(
    fixed_image: np.ndarray,
    moving_image: np.ndarray,
    interpolator=sitk.sitkLinear,
    iterations: int = 5000,
    converge_min: float = 1e-8,
):
    """
    Register two images using a rigid body transformation. This methods applies
    a Euler 3D transformation in order to register the moving image to the
    fixed image.

    The optimization method used is the Gradient Descent.

    Note:
        The registration process is based on the SimpleITK library. More details
        on how the registration process works can be found at: [Registration Overview](https://simpleitk.readthedocs.io/en/master/registrationOverview.html)

    Args:
        fixed_image (np.ndarray): The fixed image as the reference space.
        moving_image (np.ndarray): The moving image to be registered.
        interpolator (sitk.Interpolator, optional): The interpolation method used in the registration process. Defaults to sitk.sitkLinear.

    Raises:
        Exception: fixed_image and moving_image must be a numpy array.

    Returns:
        numpy.ndarray: The resampled image.
        numpy.ndarray: The transformation matrix.
    """

    # Check if the fixed_image is a numpy array.
    if not isinstance(fixed_image, np.ndarray) or not isinstance(
        moving_image, np.ndarray
    ):
        raise Exception('fixed_image and moving_image must be a numpy array.')

    fixed_image = sitk.GetImageFromArray(fixed_image)
    moving_image = sitk.GetImageFromArray(moving_image)

    # Create the registration method.
    registration_method = sitk.ImageRegistrationMethod()

    # Initialize the registration method.
    registration_transform = sitk.Euler3DTransform()
    initial_transform = sitk.CenteredTransformInitializer(
        fixed_image,
        moving_image,
        registration_transform,
        sitk.CenteredTransformInitializerFilter.GEOMETRY,
    )
    registration_method.SetInitialTransform(initial_transform)

    # Set the metric.
    registration_method.SetMetricAsMattesMutualInformation(
        numberOfHistogramBins=50
    )
    registration_method.SetMetricSamplingStrategy(registration_method.RANDOM)
    registration_method.SetMetricSamplingPercentage(0.01)

    # Set the optimizer.
    registration_method.SetOptimizerAsGradientDescent(
        learningRate=1.0,
        numberOfIterations=iterations,
        convergenceMinimumValue=converge_min,
        convergenceWindowSize=10,
    )
    registration_method.SetOptimizerScalesFromPhysicalShift()

    # Set the interpolator.
    registration_method.SetInterpolator(interpolator)

    # Execute the registration.
    final_transform = registration_method.Execute(fixed_image, moving_image)

    # Convert the final transform to a numpy array.
    transform_matrix = np.array(final_transform.GetMatrix()).reshape(3, 3)

    # Create a 4x4 transformation matrix.
    transformation_matrix = np.eye(4)
    transformation_matrix[:3, :3] = transform_matrix
    transformation_matrix[:3, 3] = final_transform.GetTranslation()

    # Resample the moving image.
    resampled_image = sitk.Resample(
        moving_image,
        fixed_image,
        final_transform,
        interpolator,
        0.0,
        moving_image.GetPixelID(),
    )

    resampled_image = sitk.GetArrayFromImage(resampled_image)
    return resampled_image, transformation_matrix


# def affine_registration(fixed_image: np.ndarray, moving_image: np.ndarray, interpolator=sitk.sitkLinear, iterations: int = 5000, converge_min: float = 1e-8):

#     # Check if the fixed_image is a numpy array.
#     if not isinstance(fixed_image, np.ndarray) or not isinstance(moving_image, np.ndarray):
#         raise Exception('fixed_image and moving_image must be a numpy array.')

#     fixed_image = sitk.GetImageFromArray(fixed_image)
#     moving_image = sitk.GetImageFromArray(moving_image)

#     # Create the registration method.
#     registration_method = sitk.ImageRegistrationMethod()

#     # Initialize the registration method.
#     registration_transform = sitk.AffineTransform(3)
#     initial_transform = sitk.CenteredTransformInitializer(fixed_image, moving_image, registration_transform,
#                                                           sitk.CenteredTransformInitializerFilter.GEOMETRY)
#     registration_method.SetInitialTransform(initial_transform)

#     # Set the metric.
#     registration_method.SetMetricAsMattesMutualInformation(numberOfHistogramBins=50)
#     registration_method.SetMetricSamplingStrategy(registration_method.RANDOM)
#     registration_method.SetMetricSamplingPercentage(0.01)

#     # Set the optimizer.
#     registration_method.SetOptimizerAsGradientDescent(learningRate=1.0, numberOfIterations=iterations,
#                                                       convergenceMinimumValue=converge_min, convergenceWindowSize=10)
#     registration_method.SetOptimizerScalesFromPhysicalShift()

#     # Set the interpolator.
#     registration_method.SetInterpolator(interpolator)

#     # Execute the registration.
#     final_transform = registration_method.Execute(fixed_image, moving_image)

#     # Convert the final transform to a numpy array.
#     transformation_matrix = np.array(final_transform.GetMatrix()).reshape(3, 3)

#     # Resample the moving image.
#     resampled_image = sitk.Resample(moving_image, fixed_image, final_transform, interpolator, 0.0,
#                                     moving_image.GetPixelID())

#     resampled_image = sitk.GetArrayFromImage(resampled_image)
#     return resampled_image, transformation_matrix



"""
The MRI parameters are defined here.

The paper references is listed below, which can be used to get more information
about some measures:

[1] Leonie Petitclerc, et al. "Ultra-long-TE arterial spin labeling reveals
rapid and brain-wide blood-to-CSF water transport in humans", Neuroimage
(2021). DOI: 10.1016/j.neuroimage.2021.118755

[2] R B Buxton, et al. "A general kinetic model for quantitative perfusion
imaging with arterial spin labeling". Magn Reson Med (1998). PMID: 9727941
DOI: 10.1002/mrm.1910400308
"""


class MRIParameters:
    def __init__(self) -> None:
        """Creates the basic MRIParameters object to define the main MRI
        constants and values for ASL processing

        To see all the parameters listed in the class, one can call print the
        values using the default `print()` function
        """
        self.T1bl = 1650.0   # T1 relaxation for the blood [1]
        self.T1csf = 1400.0   # T1 relaxation for the CSF [1] Paper Ultralong TE: T1csf = 4300!!!

        self.T2bl = 165.0  # T2 relaxation for the blood [1] PAPER CITE 150!!!!
        self.T2gm = 75.0   # T2 relaxation for the GM [1] PAPER CITE 60
        self.T2csf = 1500   # T2 relaxation for the CSF [1]

        # MRI constants
        self.Alpha = 0.85   # RF labeling efficiency
        self.Lambda = 0.98   # Blood-brain partition coefficient [1]

    def set_constant(self, value: float, param: str):
        """Set a different value for a parameter defined in the MRIParameter
        class.

        Args:
            value (float): The value to be assumed in the parameter
            param (str): The parameter that will receive the new value

        Raises:
            AttributeError: The parameter type must be already defined in the
            MRIParameters class.
        """
        if param == 'T1bl':
            self.T1bl = value
        elif param == 'T1csf':
            self.T1csf = value
        elif param == 'T2bl':
            self.T2bl = value
        elif param == 'T2gm':
            self.T2gm = value
        elif param == 'T2csf':
            self.T2csf = value
        elif param == 'Alpha':
            self.Alpha = value
        elif param == 'Lambda':
            self.Lambda = value
        else:
            raise AttributeError(
                f'Constant type {param} is not valid. Choose in the list available in the MRIParameter class.'
            )

    def get_constant(self, param: str) -> float:
        """Collect a parameter value from a defined type

        Args:
            param (str): The specific parameter that should return the storage
            value.

        Raises:
            AttributeError: The parameter type must be already defined in the
            MRIParameters class.

        Returns:
            float: The parameter value storage in the object instance
        """
        if param == 'T1bl':
            return self.T1bl
        elif param == 'T1csf':
            return self.T1csf
        elif param == 'T2bl':
            return self.T2bl
        elif param == 'T2gm':
            return self.T2gm
        elif param == 'T2csf':
            return self.T2csf
        elif param == 'Alpha':
            return self.Alpha
        elif param == 'Lambda':
            return self.Lambda
        else:
            raise AttributeError(
                f'Constant type {param} is not valid. Choose in the list available in the MRIParameter class.'
            )



import warnings
from multiprocessing import Array, Pool, cpu_count

import numpy as np
import SimpleITK as sitk
from rich import print
from rich.progress import Progress
from scipy.optimize import curve_fit

from asltk.asldata import ASLData
from asltk.aux_methods import _apply_smoothing_to_maps, _check_mask_values
from asltk.models.signal_dynamic import asl_model_multi_dw
from asltk.mri_parameters import MRIParameters
from asltk.reconstruction import CBFMapping

# Global variables to assist multi cpu threading
cbf_map = None
att_map = None
brain_mask = None
asl_data = None
ld_arr = None
pld_arr = None
te_arr = None
tblgm_map = None
t2bl = None
t2gm = None


class MultiDW_ASLMapping(MRIParameters):
    def __init__(self, asl_data: ASLData):
        """Multi-Diffusion-Weighted ASL mapping constructor for advanced perfusion analysis.

        MultiDW_ASLMapping enables sophisticated ASL analysis by incorporating multiple
        diffusion weightings (b-values) to separate intravascular and tissue
        compartments. This approach provides enhanced characterization of perfusion
        and can help differentiate between different vascular compartments.

        The class implements diffusion-weighted ASL analysis that can distinguish:
        - Fast-flowing blood (intravascular component)
        - Slow-flowing blood and tissue perfusion
        - Apparent diffusion coefficients for each compartment
        - Water exchange parameters between compartments

        Notes:
            The ASLData object must contain `dw_values` - a list of diffusion
            b-values used during ASL acquisition. These b-values are essential
            for the multi-compartment diffusion model fitting.

        Examples:
            Basic multi-DW ASL mapping setup:
            >>> from asltk.asldata import ASLData
            >>> from asltk.reconstruction import MultiDW_ASLMapping
            >>> # Create ASL data with diffusion weighting
            >>> asl_data = ASLData(
            ...     pcasl='./tests/files/pcasl_mdw.nii.gz',
            ...     m0='./tests/files/m0.nii.gz',
            ...     dw_values=[0, 50, 100, 200],    # b-values in s/mm²
            ...     ld_values=[1.8, 1.8, 1.8, 1.8],
            ...     pld_values=[0.8, 1.8, 2.8, 3.8]
            ... )
            >>> mdw_mapper = MultiDW_ASLMapping(asl_data)

            Access diffusion-related maps (after processing):
            >>> # These maps will be populated after create_map() is called
            >>> # A1: Signal amplitude for compartment 1
            >>> # D1: Apparent diffusion coefficient for compartment 1
            >>> # A2: Signal amplitude for compartment 2
            >>> # D2: Apparent diffusion coefficient for compartment 2
            >>> # kw: Water exchange parameter

        Args:
            asl_data (ASLData): The ASL data object containing multi-DW acquisition.
                Must include dw_values (b-values), ld_values, and pld_values.

        Raises:
            ValueError: If ASLData object lacks required DW values for
                diffusion-weighted analysis.

        See Also:
            CBFMapping: For basic CBF/ATT mapping without diffusion weighting
            MultiTE_ASLMapping: For multi-echo ASL analysis
        """
        super().__init__()
        self._asl_data = asl_data
        self._basic_maps = CBFMapping(asl_data)
        if self._asl_data.get_dw() is None:
            raise ValueError(
                'ASLData is incomplete. MultiDW_ASLMapping need a list of DW values.'
            )

        self._brain_mask = np.ones(self._asl_data('m0').shape)
        self._cbf_map = np.zeros(self._asl_data('m0').shape)
        self._att_map = np.zeros(self._asl_data('m0').shape)

        self._b_values = self._asl_data.get_dw()
        # self._A1 = np.zeros(tuple([len(self._b_values)]) + self._asl_data('m0').shape)
        self._A1 = np.zeros(self._asl_data('m0').shape)
        # self._D1 = np.zeros(tuple([1]) +self._asl_data('m0').shape)
        self._D1 = np.zeros(self._asl_data('m0').shape)
        self._A2 = np.zeros(self._asl_data('m0').shape)
        # self._A2 = np.zeros(tuple([len(self._b_values)])  + self._asl_data('m0').shape)
        # self._D2 = np.zeros(tuple([1]) +self._asl_data('m0').shape)
        self._D2 = np.zeros(self._asl_data('m0').shape)
        self._kw = np.zeros(self._asl_data('m0').shape)

    def set_brain_mask(self, brain_mask: np.ndarray, label: int = 1):
        """Set brain mask for MultiDW-ASL processing (strongly recommended).

        A brain mask is especially important for multi-diffusion-weighted ASL
        processing as it significantly reduces computation time by limiting
        the intensive voxel-wise fitting to brain tissue regions only.

        Without a brain mask, processing time can be prohibitively long (hours)
        for whole-volume analysis. A proper brain mask can reduce processing
        time by 5-10x while maintaining analysis quality.

        Args:
            brain_mask (np.ndarray): The image representing the brain mask.
                Must match the spatial dimensions of the M0 image.
            label (int, optional): The label value used to define brain tissue.
                Defaults to 1. Voxels with this value will be processed.

        Examples:
            Set a brain mask for efficient processing:
            >>> from asltk.asldata import ASLData
            >>> from asltk.reconstruction import MultiDW_ASLMapping
            >>> import numpy as np
            >>> asl_data = ASLData(
            ...     pcasl='./tests/files/pcasl_mdw.nii.gz',
            ...     m0='./tests/files/m0.nii.gz',
            ...     dw_values=[0, 50, 100], ld_values=[1.8]*3, pld_values=[1.8]*3
            ... )
            >>> mdw_mapper = MultiDW_ASLMapping(asl_data)
            >>> # Create conservative brain mask (center region only)
            >>> mask_shape = asl_data('m0').shape
            >>> brain_mask = np.zeros(mask_shape)
            >>> brain_mask[1:4, 5:30, 5:30] = 1  # Conservative brain region
            >>> mdw_mapper.set_brain_mask(brain_mask)

        Note:
            For multi-DW ASL, consider using a more conservative (smaller) brain
            mask initially to test parameters and processing time, then expand
            to full brain analysis once satisfied with results.
        """
        _check_mask_values(brain_mask, label, self._asl_data('m0').shape)

        binary_mask = (brain_mask == label).astype(np.uint8) * label
        self._brain_mask = binary_mask

    def get_brain_mask(self):
        """Get the brain mask image

        Returns:
            (np.ndarray): The brain mask image
        """
        return self._brain_mask

    def set_cbf_map(self, cbf_map: np.ndarray):
        """Set the CBF map to the MultiDW_ASLMapping object.

        Note:
            The CBF maps must have the original scale in order to calculate the
            T1blGM map correclty. Hence, if the CBF map was made using
            CBFMapping class, one can use the 'cbf' output.

        Args:
            cbf_map (np.ndarray): The CBF map that is set in the MultiDW_ASLMapping object
        """
        self._cbf_map = cbf_map

    def get_cbf_map(self) -> np.ndarray:
        """Get the CBF map storaged at the MultiDW_ASLMapping object

        Returns:
            (np.ndarray): The CBF map that is storaged in the
            MultiDW_ASLMapping object
        """
        return self._cbf_map

    def set_att_map(self, att_map: np.ndarray):
        """Set the ATT map to the MultiDW_ASLMapping object.

        Args:
            att_map (np.ndarray): The ATT map that is set in the MultiDW_ASLMapping object
        """
        self._att_map = att_map

    def get_att_map(self):
        """Get the ATT map storaged at the MultiDW_ASLMapping object

        Returns:
            (np.ndarray): _description_
        """
        return self._att_map

    def create_map(
        self,
        lb: list = [0.0, 0.0, 0.0, 0.0],
        ub: list = [np.inf, np.inf, np.inf, np.inf],
        par0: list = [0.5, 0.000005, 0.5, 0.000005],
        smoothing=None,
        smoothing_params=None,
    ):
        """Create multi-diffusion-weighted ASL maps for compartment analysis.

        This method performs advanced diffusion-weighted ASL analysis to generate
        multi-compartment perfusion maps. The analysis uses multiple b-values to
        separate fast-flowing intravascular and slower tissue perfusion components.

        The method fits a bi-exponential diffusion model to estimate:
        - Signal amplitudes and apparent diffusion coefficients for two compartments
        - Water exchange parameters between vascular and tissue compartments
        - Enhanced CBF characterization with compartment specificity

        Note:
            The CBF and ATT maps can be provided before calling this method using
            set_cbf_map() and set_att_map() methods. If not provided, basic maps
            are automatically calculated using the CBFMapping class.

        Warning:
            This method is computationally intensive as it performs voxel-wise
            non-linear fitting without parallel processing. Consider using a brain
            mask to limit processing to relevant tissue areas.

        Args:
            lb (list, optional): Lower bounds for [A1, D1, A2, D2] parameters.
                Defaults to [0.0, 0.0, 0.0, 0.0]. All parameters should be non-negative.
                - A1, A2: Signal amplitudes (relative units)
                - D1, D2: Apparent diffusion coefficients (mm²/s)
            ub (list, optional): Upper bounds for [A1, D1, A2, D2] parameters.
                Defaults to [np.inf, np.inf, np.inf, np.inf].
            par0 (list, optional): Initial guess for [A1, D1, A2, D2] parameters.
                Defaults to [0.5, 0.000005, 0.5, 0.000005].
                - A1, A2: Typical values 0.1-1.0 (relative amplitudes)
                - D1, D2: Typical values 1e-6 to 1e-3 mm²/s
            smoothing (str, optional): Type of spatial smoothing filter to apply.
                Options: None (default, no smoothing), 'gaussian', 'median'.
                Smoothing is applied to all output maps after reconstruction.
            smoothing_params (dict, optional): Parameters for the smoothing filter.
                For 'gaussian': {'sigma': float} (default: 1.0)
                For 'median': {'size': int} (default: 3)

        Returns:
            dict: Dictionary containing diffusion-weighted ASL maps:
                - 'cbf': Basic CBF map in original units (numpy.ndarray)
                - 'cbf_norm': Normalized CBF in mL/100g/min (numpy.ndarray)
                - 'att': Arterial transit time in ms (numpy.ndarray)
                - 'A1': Signal amplitude for compartment 1 (numpy.ndarray)
                - 'D1': Apparent diffusion coefficient for compartment 1 in mm²/s (numpy.ndarray)
                - 'A2': Signal amplitude for compartment 2 (numpy.ndarray)
                - 'D2': Apparent diffusion coefficient for compartment 2 in mm²/s (numpy.ndarray)
                - 'kw': Water exchange parameter (numpy.ndarray)
                All maps are smoothed if smoothing is enabled.

        Examples:
            Basic multi-DW ASL analysis:
            >>> from asltk.asldata import ASLData
            >>> from asltk.reconstruction import MultiDW_ASLMapping
            >>> import numpy as np
            >>> # Load multi-DW ASL data
            >>> asl_data = ASLData(
            ...     pcasl='./tests/files/pcasl_mdw.nii.gz',
            ...     m0='./tests/files/m0.nii.gz',
            ...     dw_values=[0, 50, 100, 200],  # b-values in s/mm²
            ...     ld_values=[1.8, 1.8, 1.8, 1.8],
            ...     pld_values=[0.8, 1.8, 2.8, 3.8]
            ... )
            >>> mdw_mapper = MultiDW_ASLMapping(asl_data)
            >>> # Set brain mask for faster processing (recommended)
            >>> brain_mask = np.ones(asl_data('m0').shape)
            >>> brain_mask[0:2, :, :] = 0  # Remove some background slices
            >>> mdw_mapper.set_brain_mask(brain_mask)
            >>> # Generate all maps (may take several minutes)
            >>> results = mdw_mapper.create_map() # doctest: +SKIP

            Custom parameters for specific tissue analysis:
            >>> # For analyzing fast vs slow perfusion components
            >>> results = mdw_mapper.create_map(
            ...     lb=[0.1, 1e-6, 0.1, 1e-7],      # Minimum realistic values
            ...     ub=[2.0, 1e-3, 2.0, 1e-4],      # Maximum realistic values
            ...     par0=[0.8, 5e-5, 0.3, 1e-5]     # Initial guesses
            ... ) # doctest: +SKIP

        Note:
            Processing time scales with brain mask size. For a full brain analysis,
            expect processing times of 30+ minutes depending on data size and
            hardware capabilities.

        See Also:
            set_cbf_map(): Provide pre-computed CBF map
            set_att_map(): Provide pre-computed ATT map
            CBFMapping: For basic CBF/ATT mapping
        """
        self._basic_maps.set_brain_mask(self._brain_mask)

        basic_maps = {'cbf': self._cbf_map, 'att': self._att_map}
        if np.mean(self._cbf_map) == 0 or np.mean(self._att_map) == 0:
            # If the CBF/ATT maps are zero (empty), then a new one is created
            print(
                '[blue][INFO] The CBF/ATT map were not provided. Creating these maps before next step...'
            )   # pragma: no cover
            basic_maps = self._basic_maps.create_map()   # pragma: no cover
            self._cbf_map = basic_maps['cbf']   # pragma: no cover
            self._att_map = basic_maps['att']   # pragma: no cover

        x_axis = self._asl_data('m0').shape[2]   # height
        y_axis = self._asl_data('m0').shape[1]   # width
        z_axis = self._asl_data('m0').shape[0]   # depth

        # TODO Fix
        print('multiDW-ASL processing...')
        for i in range(x_axis):
            for j in range(y_axis):
                for k in range(z_axis):
                    if self._brain_mask[k, j, i] != 0:
                        # Calculates the diffusion components for (A1, D1), (A2, D2)
                        def mod_diff(Xdata, par1, par2, par3, par4):
                            return asl_model_multi_dw(
                                b_values=Xdata,
                                A1=par1,
                                D1=par2,
                                A2=par3,
                                D2=par4,
                            )

                        # M(t,b)/M(t,0)
                        Ydata = (
                            self._asl_data('pcasl')[:, :, k, j, i]
                            .reshape(
                                (
                                    len(self._asl_data.get_ld())
                                    * len(self._asl_data.get_dw()),
                                    1,
                                )
                            )
                            .flatten()
                            / self._asl_data('m0')[k, j, i]
                        )

                        try:
                            # Xdata = self._b_values
                            Xdata = self._create_x_data(
                                self._asl_data.get_ld(),
                                self._asl_data.get_pld(),
                                self._asl_data.get_dw(),
                            )

                            par_fit, _ = curve_fit(
                                mod_diff,
                                Xdata[:, 2],
                                Ydata,
                                p0=par0,
                                bounds=(lb, ub),
                            )
                            self._A1[k, j, i] = par_fit[0]
                            self._D1[k, j, i] = par_fit[1]
                            self._A2[k, j, i] = par_fit[2]
                            self._D2[k, j, i] = par_fit[3]
                        except RuntimeError:
                            self._A1[k, j, i] = 0
                            self._D1[k, j, i] = 0
                            self._A2[k, j, i] = 0
                            self._D2[k, j, i] = 0

                        # Calculates the Mc fitting to alpha = kw + T1blood
                        m0_px = self._asl_data('m0')[k, j, i]

                        # def mod_2comp(Xdata, par1):
                        #     ...
                        #     # return asl_model_multi_te(
                        #     #     Xdata[:, 0],
                        #     #     Xdata[:, 1],
                        #     #     Xdata[:, 2],
                        #     #     m0_px,
                        #     #     basic_maps['cbf'][k, j, i],
                        #     #     basic_maps['att'][k, j, i],
                        #     #     par1,
                        #     #     self.T2bl,
                        #     #     self.T2gm,
                        #     # )

                        # Ydata = (
                        #     self._asl_data('pcasl')[:, :, k, j, i]
                        #     .reshape(
                        #         (
                        #             len(self._asl_data.get_ld())
                        #             * len(self._asl_data.get_te()),
                        #             1,
                        #         )
                        #     )
                        #     .flatten()
                        # )

                        # try:
                        #     Xdata = self._create_x_data(
                        #         self._asl_data.get_ld(),
                        #         self._asl_data.get_pld(),
                        #         self._asl_data.get_dw(),
                        #     )
                        #     par_fit, _ = curve_fit(
                        #         mod_2comp,
                        #         Xdata,
                        #         Ydata,
                        #         p0=par0,
                        #         bounds=(lb, ub),
                        #     )
                        #     self._kw[k, j, i] = par_fit[0]
                        # except RuntimeError:
                        #     self._kw[k, j, i] = 0.0

        # # Adjusting output image boundaries
        # self._kw = self._adjust_image_limits(self._kw, par0[0])

        # Create output maps dictionary
        output_maps = {
            'cbf': self._cbf_map,
            'cbf_norm': self._cbf_map * (60 * 60 * 1000),
            'att': self._att_map,
            'a1': self._A1,
            'd1': self._D1,
            'a2': self._A2,
            'd2': self._D2,
            'kw': self._kw,
        }

        # Apply smoothing if requested
        return _apply_smoothing_to_maps(
            output_maps, smoothing, smoothing_params
        )

    def _create_x_data(self, ld, pld, dw):
        # array for the x values, assuming an arbitrary size based on the PLD
        # and TE vector size
        Xdata = np.zeros((len(pld) * len(dw), 3))

        count = 0
        for i in range(len(pld)):
            for j in range(len(dw)):
                Xdata[count] = [ld[i], pld[i], dw[j]]
                count += 1

        return Xdata



import fnmatch
import os

import dill
import numpy as np
import SimpleITK as sitk
from bids import BIDSLayout

from asltk import AVAILABLE_IMAGE_FORMATS, BIDS_IMAGE_FORMATS


def _check_input_path(full_path: str):
    if not os.path.exists(full_path):
        raise FileNotFoundError(f'The file {full_path} does not exist.')


def _get_file_from_folder_layout(
    full_path: str,
    subject: str = None,
    session: str = None,
    modality: str = None,
    suffix: str = None,
):
    selected_file = None
    layout = BIDSLayout(full_path)
    if all(param is None for param in [subject, session, modality, suffix]):
        for root, _, files in os.walk(full_path):
            for file in files:
                if '_asl' in file and file.endswith(BIDS_IMAGE_FORMATS):
                    selected_file = os.path.join(root, file)
    else:
        layout_files = layout.files.keys()
        matching_files = []
        for f in layout_files:
            search_pattern = ''
            if subject:
                search_pattern = f'*sub-*{subject}*'
            if session:
                search_pattern += search_pattern + f'*ses-*{session}'
            if modality:
                search_pattern += search_pattern + f'*{modality}*'
            if suffix:
                search_pattern += search_pattern + f'*{suffix}*'

            if fnmatch.fnmatch(f, search_pattern) and f.endswith(
                BIDS_IMAGE_FORMATS
            ):
                matching_files.append(f)

        if not matching_files:
            raise FileNotFoundError(
                f'ASL image file is missing for subject {subject} in directory {full_path}'
            )
        selected_file = matching_files[0]

    return selected_file


def load_image(
    full_path: str,
    subject: str = None,
    session: str = None,
    modality: str = None,
    suffix: str = None,
):
    """
    Load an image file from a BIDS directory or file using the SimpleITK API.

    The output is always a numpy array, converted from the SimpleITK image object.

    Supported image formats include: .nii, .nii.gz, .nrrd, .mha, .tif, and other formats supported by SimpleITK.

    Note:
        - The default values for `modality` and `suffix` are None. If not provided, the function will search for the first matching ASL image in the directory.
        - If `full_path` is a file, it is loaded directly. If it is a directory, the function searches for a BIDS-compliant image using the provided parameters.
        - If both a file and a BIDS directory are provided, the file takes precedence.

    Tip:
        To validate your BIDS structure, use the `bids-validator` tool: https://bids-standard.github.io/bids-validator/
        For more details about ASL BIDS structure, see: https://bids-specification.readthedocs.io/en/latest

    Note:
        The image file is assumed to be an ASL subtract image (control-label). If not, use helper functions in `asltk.utils` to create one.

    Args:
        full_path (str): Path to the image file or BIDS directory.
        subject (str, optional): Subject identifier. Defaults to None.
        session (str, optional): Session identifier. Defaults to None.
        modality (str, optional): Modality folder name. Defaults to None.
        suffix (str, optional): Suffix of the file to load. Defaults to None.

    Examples:
        Load a single image file directly:
        >>> data = load_image("./tests/files/pcasl_mte.nii.gz")
        >>> type(data)
        <class 'numpy.ndarray'>
        >>> data.shape  # Example: 5D ASL data
        (8, 7, 5, 35, 35)

        Load M0 reference image:
        >>> m0_data = load_image("./tests/files/m0.nii.gz")
        >>> m0_data.shape  # Example: 3D reference image
        (5, 35, 35)

        Load from BIDS directory (automatic detection):
        >>> data = load_image("./tests/files/bids-example/asl001")
        >>> type(data)
        <class 'numpy.ndarray'>

        Load specific BIDS data with detailed parameters:
        >>> data = load_image("./tests/files/bids-example/asl001", subject='Sub103', suffix='asl')
        >>> type(data)
        <class 'numpy.ndarray'>

        # Load NRRD format
        >>> nrrd_data = load_image("./tests/files/t1-mri.nrrd")
        >>> type(nrrd_data)
        <class 'numpy.ndarray'>

    Returns:
        numpy.ndarray: The loaded image array.
    """
    _check_input_path(full_path)

    if full_path.endswith(AVAILABLE_IMAGE_FORMATS):
        # If the full path is a file, then load the image directly
        img = sitk.ReadImage(full_path)
        return sitk.GetArrayFromImage(img)

    # Check if the full path is a directory using BIDS structure
    selected_file = _get_file_from_folder_layout(
        full_path, subject, session, modality, suffix
    )

    img = sitk.ReadImage(selected_file)
    return sitk.GetArrayFromImage(img)


def _make_bids_path(
    bids_root, subject, session=None, suffix='asl', extension='.nii.gz'
):
    subj_dir = f'sub-{subject}'
    ses_dir = f'ses-{session}' if session else None
    modality_dir = 'asl'

    if ses_dir:
        out_dir = os.path.join(bids_root, subj_dir, ses_dir, modality_dir)
    else:
        out_dir = os.path.join(bids_root, subj_dir, modality_dir)

    os.makedirs(out_dir, exist_ok=True)

    filename = f'sub-{subject}'
    if session:
        filename += f'_ses-{session}'
    filename += f'_{suffix}{extension}'

    return os.path.join(out_dir, filename)


def save_image(
    img: np.ndarray,
    full_path: str = None,
    *,
    bids_root: str = None,
    subject: str = None,
    session: str = None,
):
    """
    Save an image to a file path using SimpleITK.

    All available image formats provided in the SimpleITK API can be used here. Supported formats include: .nii, .nii.gz, .nrrd, .mha, .tif, and others.

    Note:
        If the file extension is not recognized by SimpleITK, an error will be raised.
        The input array should be 2D, 3D, or 4D. For 4D arrays, only the first volume may be saved unless handled explicitly.

    Args:
        img (np.ndarray): The image array to be saved. Can be 2D, 3D, or 4D.
        full_path (str): Full absolute path with image file name provided.
        bids_root (str): Optional BIDS root directory to save in BIDS structure.
        subject (str): Subject ID for BIDS saving.
        session (str): Optional session ID for BIDS saving.

    Examples:
        Save an image using a direct file path:
        >>> import tempfile
        >>> import numpy as np
        >>> img = np.random.rand(10, 10, 10)
        >>> with tempfile.NamedTemporaryFile(suffix='.nii.gz', delete=False) as f:
        ...     save_image(img, f.name)

        Save an image using BIDS structure:
        >>> import tempfile
        >>> img = np.random.rand(10, 10, 10)
        >>> with tempfile.TemporaryDirectory() as temp_dir:
        ...     save_image(img, bids_root=temp_dir, subject='001', session='01')

        Save processed ASL results:
        >>> from asltk.asldata import ASLData
        >>> asl_data = ASLData(pcasl='./tests/files/pcasl_mte.nii.gz', m0='./tests/files/m0.nii.gz')
        >>> processed_img = asl_data('pcasl')[0]  # Get first volume
        >>> import tempfile
        >>> with tempfile.NamedTemporaryFile(suffix='.nii.gz', delete=False) as f:
        ...     save_image(processed_img, f.name)

    Raises:
        ValueError: If neither full_path nor (bids_root + subject) are provided.
        RuntimeError: If the file extension is not recognized by SimpleITK.
    """
    if bids_root and subject:
        full_path = _make_bids_path(bids_root, subject, session)

    if not full_path:
        raise ValueError(
            'Either full_path or bids_root + subject must be provided.'
        )

    sitk_img = sitk.GetImageFromArray(img)
    sitk.WriteImage(sitk_img, full_path)


def save_asl_data(
    asldata,
    fullpath: str = None,
    *,
    bids_root: str = None,
    subject: str = None,
    session: str = None,
):
    """
    Save ASL data to a pickle file using dill serialization.

    This method saves the ASL data to a pickle file using the dill library. All
    the ASL data will be saved in a single file. After the file is saved, it
    can be loaded using the `load_asl_data` method.

    Note:
        This method only accepts the ASLData object as input. If you want to
        save an image, use the `save_image` method.
        The file is serialized with dill, which supports more Python objects than standard pickle. However, files saved with dill may not be compatible with standard pickle, especially for custom classes.

    Parameters:
        asldata : ASLData
            The ASL data to be saved. This can be any Python object that is serializable by dill.
        fullpath : str
            The full path where the pickle file will be saved. The filename must end with '.pkl'.

    Examples:
        >>> from asltk.asldata import ASLData
        >>> asldata = ASLData(pcasl='./tests/files/pcasl_mte.nii.gz', m0='./tests/files/m0.nii.gz',ld_values=[1.8, 1.8, 1.8], pld_values=[1.8, 1.8, 1.8], te_values=[1.8, 1.8, 1.8])
        >>> import tempfile
        >>> with tempfile.NamedTemporaryFile(delete=False, suffix='.pkl') as temp_file:
        ...     temp_file_path = temp_file.name
        >>> save_asl_data(asldata, temp_file_path)

    Raises:
        ValueError: If the provided filename does not end with '.pkl'.
    """
    if bids_root and subject:
        fullpath = _make_bids_path(
            bids_root, subject, session, suffix='asl', extension='.pkl'
        )

    if not fullpath:
        raise ValueError(
            'Either fullpath or bids_root + subject must be provided.'
        )

    if not fullpath.endswith('.pkl'):
        raise ValueError('Filename must be a pickle file (.pkl)')

    dill.dump(asldata, open(fullpath, 'wb'))


def load_asl_data(fullpath: str):
    """
    Load ASL data from a specified file path to an ASLData object previously saved on disk.

    This function uses the `dill` library to load and deserialize data from a
    file. Therefore, the file must have been saved using the `save_asl_data` function.

    Note:
        The file must have been saved with dill. Files saved with dill may not be compatible with standard pickle, especially for custom classes.

    Parameters:
        fullpath (str): The full path to the file containing the serialized ASL data.

    Returns:
        ASLData: The deserialized ASL data object from the file.

    Examples:
        >>> from asltk.asldata import ASLData
        >>> asldata = ASLData(pcasl='./tests/files/pcasl_mte.nii.gz', m0='./tests/files/m0.nii.gz',ld_values=[1.8, 1.8, 1.8], pld_values=[1.8, 1.8, 1.8], te_values=[1.8, 1.8, 1.8])
        >>> import tempfile
        >>> with tempfile.NamedTemporaryFile(delete=False, suffix='.pkl') as temp_file:
        ...     temp_file_path = temp_file.name
        >>> save_asl_data(asldata, temp_file_path)
        >>> loaded_asldata = load_asl_data(temp_file_path)
        >>> loaded_asldata.get_ld()
        [1.8, 1.8, 1.8]
        >>> loaded_asldata('pcasl').shape
        (8, 7, 5, 35, 35)
    """
    _check_input_path(fullpath)
    return dill.load(open(fullpath, 'rb'))


def collect_data_volumes(data: np.ndarray):
    """
    Collect the data volumes from a higher dimension array.

    This method is used to collect the data volumes from a higher dimension
    array. The method works with 4D or 5D arrays, where the volumes are
    separated along the higher dimensions. The method will collect the volumes
    and return a list of 3D arrays.

    Note:
        If the input is already 3D, the function returns a list with a single volume (the input itself).
        For 4D or 5D arrays with singleton dimensions, the output list will contain all possible 3D volumes, even if some are identical or empty.

    The method is useful when you want to:
    - Apply filters to each volume separately
    - Process multi-echo or multi-b-value ASL data
    - Separate time series data into individual volumes
    - Prepare data for volume-wise analysis

    Args:
        data (np.ndarray): The data to be separated. Must be at least 3D.

    Returns:
        tuple: A tuple containing:
            - list: A list of 3D arrays, each one representing a volume.
            - tuple: The original shape of the data.

    Examples:
        Separate 4D ASL data into individual volumes:
        >>> from asltk.asldata import ASLData
        >>> asl_data = ASLData(pcasl='./tests/files/pcasl_mte.nii.gz', m0='./tests/files/m0.nii.gz')
        >>> pcasl_4d = asl_data('pcasl')
        >>> volumes, original_shape = collect_data_volumes(pcasl_4d)
        >>> len(volumes)  # Number of volumes
        56
        >>> volumes[0].shape  # Shape of each volume
        (5, 35, 35)
        >>> original_shape  # Original 5D shape
        (8, 7, 5, 35, 35)

        Process each volume separately:
        >>> volumes, shape = collect_data_volumes(pcasl_4d)
        >>> # Apply processing to first volume
        >>> processed_vol = volumes[0] * 1.5  # Example processing
        >>> processed_vol.shape
        (5, 35, 35)

        Work with 3D data (single volume):
        >>> import numpy as np
        >>> single_vol = np.random.rand(10, 20, 30)
        >>> volumes, shape = collect_data_volumes(single_vol)
        >>> len(volumes)
        1
        >>> volumes[0].shape
        (10, 20, 30)

        Edge case: 4D with singleton dimension
        >>> arr = np.random.rand(1, 10, 20, 30)
        >>> vols, shape = collect_data_volumes(arr)
        >>> len(vols)
        1
        >>> vols[0].shape
        (10, 20, 30)

    Raises:
        TypeError: If data is not a numpy array.
        ValueError: If data has less than 3 dimensions.
    """
    if not isinstance(data, np.ndarray):
        raise TypeError('data is not a numpy array.')

    if data.ndim < 3:
        raise ValueError('data is a 3D volume or higher dimensions')

    volumes = []
    # Calculate the number of volumes by multiplying all dimensions except the last three
    num_volumes = int(np.prod(data.shape[:-3]))
    reshaped_data = data.reshape((int(num_volumes),) + data.shape[-3:])
    for i in range(num_volumes):
        volumes.append(reshaped_data[i])

    return volumes, data.shape



import argparse
import os
from functools import *

import numpy as np
from rich import print

from asltk.asldata import ASLData
from asltk.logging_config import (
    configure_for_scripts,
    get_logger,
    log_processing_step,
)
from asltk.reconstruction import T2Scalar_ASLMapping
from asltk.utils import load_image, save_image

parser = argparse.ArgumentParser(
    prog='T2 Scalar Mapping from ASL Multi-TE ASLData',
    description='Python script to calculate the T2 scalar map from the ASL Multi-TE ASLData.',
)
parser._action_groups.pop()
required = parser.add_argument_group(title='Required parameters')
optional = parser.add_argument_group(title='Optional parameters')

required.add_argument(
    'pcasl',
    type=str,
    help='ASL raw data obtained from the MRI scanner. This must be the multi-TE ASL MRI acquisition protocol.',
)
required.add_argument(
    'm0', type=str, help='M0 image reference used to calculate the ASL signal.'
)
optional.add_argument(
    'mask',
    type=str,
    nargs='?',
    default='',
    help='Image mask defining the ROI where the calculations must be done. Any pixel value different from zero will be assumed as the ROI area. Outside the mask (value=0) will be ignored. If not provided, the entire image space will be calculated.',
)
required.add_argument(
    'out_folder',
    type=str,
    nargs='?',
    default=os.path.expanduser('~'),
    help='The output folder that is the reference to save all the output images in the script.',
)
optional.add_argument(
    '--pld',
    type=str,
    nargs='+',
    required=False,
    default=[170.0, 270.0, 370.0, 520.0, 670.0, 1070.0, 1870.0],
    help='Posts Labeling Delay (PLD) trend, arranged in a sequence of float numbers. If not passed, the default values will be used.',
)
optional.add_argument(
    '--ld',
    type=str,
    nargs='+',
    required=False,
    default=[100.0, 100.0, 150.0, 150.0, 400.0, 800.0, 1800.0],
    help='Labeling Duration trend (LD), arranged in a sequence of float numbers. If not passed, the default values will be used.',
)
optional.add_argument(
    '--te',
    type=str,
    nargs='+',
    required=False,
    default=[13.56, 67.82, 122.08, 176.33, 230.59, 284.84, 339.1, 393.36],
    help='Time of Echos (TE), arranged in a sequence of float numbers. If not passed, the default values will be used.',
)
optional.add_argument(
    '--verbose',
    action='store_true',
    help='Show more details thoughout the processing.',
)
optional.add_argument(
    '--file_fmt',
    type=str,
    nargs='?',
    default='nii',
    help='The file format that will be used to save the output images. It is not allowed image compression (ex: .gz, .zip, etc). Default is nii, but it can be choosen: mha, nrrd.',
)

args = parser.parse_args()

# Configure logging based on verbose flag
configure_for_scripts(verbose=args.verbose)
logger = get_logger('t2_maps_script')

# Script check-up parameters
def checkUpParameters():
    is_ok = True
    # Check output folder exist
    if not (os.path.isdir(args.out_folder)):
        error_msg = f'Output folder path does not exist (path: {args.out_folder}). Please create the folder before executing the script.'
        logger.error(error_msg)
        print(error_msg)
        is_ok = False

    # Check ASL image exist
    if not (os.path.isfile(args.pcasl)):
        error_msg = f'ASL input file does not exist (file path: {args.pcasl}). Please check the input file before executing the script.'
        logger.error(error_msg)
        print(error_msg)
        is_ok = False

    # Check M0  image exist
    if not (os.path.isfile(args.m0)):
        error_msg = f'M0 input file does not exist (file path: {args.m0}). Please check the input file before executing the script.'
        logger.error(error_msg)
        print(error_msg)
        is_ok = False

    if args.file_fmt not in ('nii', 'mha', 'nrrd'):
        error_msg = f'File format is not allowed or not available. The select type is {args.file_fmt}, but options are: nii, mha or nrrd'
        logger.error(error_msg)
        print(error_msg)
        is_ok = False

    return is_ok


asl_img = load_image(args.pcasl)
m0_img = load_image(args.m0)

mask_img = np.ones(asl_img[0, 0, :, :, :].shape)
if args.mask != '':
    mask_img = load_image(args.mask)


try:
    te = [float(s) for s in args.te]
    pld = [float(s) for s in args.pld]
    ld = [float(s) for s in args.ld]
except:
    te = [float(s) for s in str(args.te[0]).split()]
    pld = [float(s) for s in str(args.pld[0]).split()]
    ld = [float(s) for s in str(args.ld[0]).split()]

if not checkUpParameters():
    raise RuntimeError(
        'One or more arguments are not well defined. Please, revise the script call.'
    )


# Step 2: Show the input information to assist manual conference
logger.info('T2 Scalar processing started')
if args.verbose:
    print(' --- Script Input Data ---')
    print('ASL file path: ' + args.pcasl)
    print('ASL image dimension: ' + str(asl_img.shape))
    print('Mask file path: ' + args.mask)
    print('Mask image dimension: ' + str(mask_img.shape))
    print('M0 file path: ' + args.m0)
    print('M0 image dimension: ' + str(m0_img.shape))
    print('PLD: ' + str(pld))
    print('LD: ' + str(ld))
    print('TE: ' + str(te))
    print('Output file format: ' + str(args.file_fmt))

# Log input parameters
logger.info(f'Input parameters - PLD: {pld}, LD: {ld}, TE: {te}')
logger.info(f'Output format: {args.file_fmt}')

log_processing_step(
    'Creating ASLData object', f'Multi-TE with {len(te)} echo times'
)
data = ASLData(
    pcasl=args.pcasl, m0=args.m0, ld_values=ld, pld_values=pld, te_values=te
)

log_processing_step('Initializing T2 Scalar mapper')
recon = T2Scalar_ASLMapping(data)
recon.set_brain_mask(mask_img)


log_processing_step(
    'Generating T2 Scalar ASL maps', 'this may take several minutes'
)
maps = recon.create_map()
logger.info('T2 Scalar ASL map generation completed successfully')

log_processing_step('Saving output maps')
save_path = args.out_folder + os.path.sep + 't2_maps.' + args.file_fmt
if args.verbose and maps['t2'] is not None:
    print('Saving T2 maps - Path: ' + save_path)
logger.info(f'Saving T2 maps to: {save_path}')
save_image(maps['t2'], save_path)

if args.verbose:
    print('Execution: ' + parser.prog + ' finished successfully!')
logger.info('T2 Scalar ASL processing completed successfully')
